<'
# Copyright (c) 2012-2014 LAAS/CNRS
# All rights reserved.
#
# Redistribution and use  in source  and binary  forms,  with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of  source  code must retain the  above copyright
#      notice and this list of conditions.
#   2. Redistributions in binary form must reproduce the above copyright
#      notice and  this list of  conditions in the  documentation and/or
#      other materials provided with the distribution.
#
#                                      Anthony Mallet on Wed Mar 14 2012
#

if {[llength $argv] != 1} { error "expected arguments: component" }
lassign $argv component

# compute handy shortcuts
set comp [$component name]
set COMP [string toupper [$component name]]

lang c
'>
#include "autoconf/acheader.h"

#include <stdio.h>

#include "h2devLib.h"
#include "posterLib.h"

#include "<"$comp">/c/client.h"
#include "<"$comp">_msglib.h"
#include "private.h"
#include "serialize.h"


/* --- local data ---------------------------------------------------------- */

<'foreach s [$component services] {'>
int	genom_<"$comp">_<"[$s name]">_encode(char *buffer, int size,
		char *dst, int maxsize);
int	genom_<"$comp">_<"[$s name]">_decode(char *buffer, int size,
		char *dst, int maxsize);
<'}'>
<'foreach p [$component ports out] {'>
int	genom_<"$comp">_<"[$p name]">_decode(char *buffer,
		<"[[$p datatype] argument reference dst]">);
<'}'>


/* <"[--- ${comp}_abort -------------------------------------------------]"> */

static void
genom_<"$comp">_client_abort_cb(genom_client h, int rqstid, int done,
                       genom_event report, void *output, const void *exdetail,
                       void *cb_data)
{
  (void)pocolibs_clean(h, rqstid);
}

genom_event
genom_<"$comp">_client_abort(genom_client h, int rqstid)
{
  int abrtid, rid;
  genom_event ex;
  int e;

  if (rqstid < 0 || rqstid >= h->nactive)
    return genom_no_such_activity(&h->context);
  if (h->active_rqst[rqstid].status == FREE)
    return genom_no_such_activity(&h->context);
  if (h->active_rqst[rqstid].status == WAITING_INTERMED_REPLY)
    return genom_syserr(&(genom_syserr_detail){.code = EAGAIN}, &h->context);

  ex = pocolibs_newrqst(h, 0, &abrtid);
  if (ex) return ex;

  e = csClientRqstSend(
    h->csid, <"$COMP">_abort_activity_RQSTID,
    (char *)&h->active_rqst[rqstid].aid, 0,
    genom_<"$comp">_abort_activity_encode, FALSE, 0, 0, &rid);
  if (e != OK) {
    genom_mwerr_detail d;
    h2getErrMsg(errnoGet(), d.what, sizeof(d.what));
    return genom_mwerr(&d, &h->context);
  }

  h->active_rqst[abrtid].info = genom_<"$comp">_client_abort_activity_info;
  h->active_rqst[abrtid].rid = rid;
  h->active_rqst[abrtid].decode = genom_<"$comp">_abort_activity_decode;
  h->active_rqst[abrtid].sentcb = NULL;
  h->active_rqst[abrtid].donecb = genom_<"$comp">_client_abort_cb;
  h->active_rqst[abrtid].cb_data = NULL;
  h->active_rqst[abrtid].status = WAITING_FINAL_REPLY;

  return genom_ok;
}

<'foreach s [$component services] {'>

/* <"[--- ${comp}_[$s name]_rqst ----------------------------------------]"> */

genom_event
genom_<"$comp">_client_<"[$s name]">_rqst(
  genom_client h,
  const struct genom_<"$comp">_<"[$s name]">_input *in,
  <"$comp">_<"[$s name]">_cb sentcb, <"$comp">_<"[$s name]">_cb donecb,
  void *cb_data,
  int *rqstid)
{
<'  set noack [catch {$s task}]'>
  genom_event ex;
  int rid;
  int e;
  assert(h);

  ex = pocolibs_newrqst(
    h, sizeof(struct genom_<"$comp">_<"[$s name]">_output), rqstid);
  if (ex) return ex;

  e = csClientRqstSend(
    h->csid, <"$COMP">_<"[$s name]">_RQSTID,
    (char *)in, 0, genom_<"$comp">_<"[$s name]">_encode,
    <"[expr ${noack} ? "FALSE" : "TRUE"]">, 0, 0, &rid);
  if (e != OK) {
    genom_mwerr_detail d;
    h2getErrMsg(errnoGet(), d.what, sizeof(d.what));
    return genom_mwerr(&d, &h->context);
  }

  genom_<"$comp">_client_<"[$s name]">_init_output(
    h->active_rqst[*rqstid].output);
  h->active_rqst[*rqstid].info =
    genom_<"$comp">_client_<"[$s name]">_info;
  h->active_rqst[*rqstid].rid = rid;
  h->active_rqst[*rqstid].decode = genom_<"$comp">_<"[$s name]">_decode;
  h->active_rqst[*rqstid].sentcb = (genom_request_cb)sentcb;
  h->active_rqst[*rqstid].donecb = (genom_request_cb)donecb;
  h->active_rqst[*rqstid].cb_data = cb_data;
<'  if {$noack} {'>
  h->active_rqst[*rqstid].status = WAITING_FINAL_REPLY;
<'  } else {'>
  h->active_rqst[*rqstid].status = WAITING_INTERMED_REPLY;
<'  }'>

  return genom_ok;
}


/* <"[--- ${comp}_[$s name]_result --------------------------------------]"> */

genom_event
genom_<"$comp">_client_<"[$s name]">_result(
  genom_client h, int rqstid, genom_event *report,
  struct genom_<"$comp">_<"[$s name]">_output **out, void **exdetail)
{
  struct genom_client_request *a;
  genom_event s;
  void *p;

  if (rqstid < 0 || rqstid >= h->nactive)
    return genom_syserr(&(genom_syserr_detail){.code = ENOENT}, &h->context);

  a = &h->active_rqst[rqstid];
  if (a->info != genom_<"$comp">_client_<"[$s name]">_info)
    return genom_syserr(&(genom_syserr_detail){.code = ENOMSG}, &h->context);

  s = pocolibs_result(h, rqstid, report, &p, exdetail);
  if (s) return s;
  if (out) *out = p;
  return genom_ok;
}

<'}'>

<'foreach p [$component ports out simple] {'>

/* <"[--- ${comp}_[$p name]_port ----------------------------------------]"> */

genom_event
genom_<"$comp">_client_<"[$p name]">_port(
  genom_client h,
  <"[[$p datatype] argument reference data]">)
{
  static POSTER_ID id;
  int searched = 0;
  char *addr;
  int s;

  do {
    if (!id) {
      char name[H2_DEV_MAX_NAME];
      int n;

      n = snprintf(name, sizeof(name), "%s/<"[$p name]">", h->instance);
      if (n <= 0 || n >= sizeof(name))
        return genom_syserr(&(genom_syserr_detail){.code = ENAMETOOLONG},
                            &h->context);

      if (posterFind(name, &id) != OK) {
        genom_mwerr_detail d;
        h2getErrMsg(errnoGet(), d.what, sizeof(d.what));
        return genom_mwerr(&d, &h->context);
      }
      searched = 1;
    }

    if (posterTake(id, POSTER_READ) != OK) {
      genom_mwerr_detail d;
      int code = errnoGet();
      if (!searched && code == S_posterLib_POSTER_CLOSED) {
        id = NULL; continue /* try to refetch poster id */;
      }
      h2getErrMsg(code, d.what, sizeof(d.what));
      return genom_mwerr(&d, &h->context);
    }
    addr = posterAddr(id);
    s = genom_<"$comp">_<"[$p name]">_decode(addr, data);
    posterGive(id);
    break;
  } while(1);

  if (s)
    return genom_syserr(&(genom_syserr_detail){.code = s}, &h->context);
  return genom_ok;
}

<'}'>
<'foreach p [$component ports out multiple] {'>

/* <"[--- ${comp}_[$p name]_port ----------------------------------------]"> */

genom_event
genom_<"$comp">_client_<"[$p name]">_port(
  genom_client h,
  const char *path,
  <"[[$p datatype] argument reference data]">)
{
  char name[H2_DEV_MAX_NAME];
  POSTER_ID id;
  char *addr;
  int s, n;

  n = snprintf(name, sizeof(name), "%s/<"[$p name]">/%s", h->instance, path);
  if (n <= 0 || n >= sizeof(name))
    return genom_syserr(&(genom_syserr_detail){.code = ENAMETOOLONG},
                        &h->context);

  if (posterFind(name, &id) != OK || posterTake(id, POSTER_READ) != OK) {
    genom_mwerr_detail d;
    h2getErrMsg(errnoGet(), d.what, sizeof(d.what));
    return genom_mwerr(&d, &h->context);
  }

  addr = posterAddr(id);
  s = genom_<"$comp">_<"[$p name]">_decode(addr, data);
  posterGive(id);

  if (s)
    return genom_syserr(&(genom_syserr_detail){.code = s}, &h->context);
  return genom_ok;
}

<'}'>
